clc;
clear all;

global G M_moon R_moon r_rif v_rif m_rif t_rif a_rif;
G = 6.67e-11;   % m^3/kg/s^2
M_moon = 7.342e22; % kg
R_moon = 1.738e6;  % m
r_rif = 15000;     % m
v_rif = 1000;       % m/s
m_rif = 1600;      % kg
t_rif = r_rif / v_rif; % s
a_rif = v_rif^2./ r_rif;

% Condizioni iniziali
y0 = [-500; 0; 15000; sqrt(G*M_moon/(R_moon + r_rif)); 0; 0; 1600]; % Dimensionale
y0_adim = y0 ./ [r_rif; r_rif; r_rif; v_rif; v_rif; v_rif; m_rif]; % Adimensionale
problems.x0 = y0_adim; % Stato iniziale del sistema

% Inizializzazione dei vincoli sul tempo finale
problems.tfMin = 25;  % Limite inferiore per tf_adim
problems.tfMax = 45;% Limite superiore per tf_adim

% Impostazione dei vincoli sui controlli 
problems.controlMin = [2]; % Limite inferiore per il controllo
problems.controlMax = [pi];  % Limite superiore per il controllo
problems.uguess(:, 1) = [2.4; 2.95]; % Stima iniziale per i controlli
problems.tguess = [0; 30]; % Stima iniziale per il tempo
problems.nodes = 100; % Numero di nodi per la discretizzazione temporale

solution = risolvi(problems);

figure(1); clf;
hold on
plot(solution.time, solution.state(:, 6), 'o'); % Stato x (punti)
hold off;
grid on; % Attiva la griglia
xlabel('Tempo Adimensionale [/]'); % Etichetta per l'asse x
ylabel('Velocità verticale [/]'); % Etichetta per l'asse y
title('Velocità verticale')

figure(2);
plot(solution.state(:, 1), solution.state(:, 3), 'o'); % Traiettoria discreta
grid on; % Attiva la griglia
xlabel('Stato x [/]'); % Etichetta per l'asse x
ylabel('Altitudine z [/]'); % Etichetta per l'asse y
title('Discesa da h_0=15km')

figure(3); clf;
plot(solution.time, solution.control, 'o'); % Controllo discreto
grid on; % Attiva la griglia
xlabel('Tempo Adimensionale [/]'); % Etichetta per l'asse x
ylabel('Angolo di spinta [rad]'); % Etichetta per l'asse y
title('Angolo di controllo');

figure(4); clf;
hold on
plot(solution.time(91:100)*t_rif, solution.state(91:100, 4).*v_rif, 'o'); % Stato x (punti)
hold off;
grid on; % Attiva la griglia
xlabel('Tempo [s]'); % Etichetta per l'asse x
ylabel('Velocità orizzontale [m/s]'); % Etichetta per l'asse y
title('Velocità verticale di atterraggio')



function solutionOut=risolvi(problem)
global t_rif
    problem.ns = length(problem.x0);
    problem.nc = length(problem.controlMin);
    problem.nuVar = problem.nc * problem.nodes; % Variabili di controllo totali
    problem.npts = linspace(0, 1, problem.nodes); % Punti normalizzati
    
    % 4. Indici delle variabili
    problem.numVar = problem.nuVar + 1; % Numero totale di variabili (controlli + tf)
    problem.nuIndex = 1:problem.nuVar; % Indici per le variabili di controllo
    problem.ntIndex = problem.nuVar + 1; % Indice per tf
    
    % 5. Interpolazione delle variabili di controllo
    uVar = interp1(problem.tguess / problem.tguess(end), problem.uguess, problem.npts, "linear");
    % 6. Inizializzazione delle variabili
    allVarGuess = [uVar(:); problem.tguess(end)]; % Insieme delle variabili iniziali
    varMin = repmat(problem.controlMin, 1, problem.nodes)'; % Limiti inferiori
    varMax = repmat(problem.controlMax, 1, problem.nodes)'; % Limiti superiori
    allVarLower = [varMin(:); problem.tfMin]; % Limiti inferiori totali
    allVarUpper = [varMax(:); problem.tfMax]; % Limiti superiori totali

    optODE = odeset('RelTol', 1e-3, 'AbsTol', 1e-3); % Opzioni per l'ODE
    problem.optODE = optODE;
    
    % 9. Opzioni per l'ottimizzazione
    options = optimoptions('fmincon', ...
            'Algorithm', 'sqp', ...                    % Cambiamo l'algoritmo a 'sqp'
            'ConstraintTolerance', 1e-3, ...           % Rilassiamo la tolleranza sui vincoli
            'StepTolerance', 1e-3, ...                 % Aumentiamo la tolleranza sui passi
            'Display', 'iter-detailed', ...            % Mostra dettagli sull'ottimizzazione ad ogni iterazione
            'MaxIterations', 2000, ...                 % Aumentiamo il numero massimo di iterazioni
            'OptimalityTolerance', 1e-3, ...           % Tolleranza per la condizione di ottimalità
            'FunctionTolerance', 1e-3);                % Tolleranza per il valore della funzione obiettivo
        % 10. Risoluzione del problema di ottimizzazione
    t1 = tic; % Inizio cronometro
    x = fmincon(@(allVar) objectivee(allVar, problem), allVarGuess, [], [], [], [], allVarLower, allVarUpper, @(allVar) constraintss(allVar, problem), options);
    solTime = toc(t1); % Tempo di soluzione
    
    % 11. Estrazione dei risultati
    controlRet = x(problem.nuIndex) % Controlli ottimizzati
    tf = x(problem.ntIndex)

    t_combined = [];
    y_combined = [];
    
    % Loop per l'integrazione su ogni intervallo tspan
    tspan=tf/problem.nodes;
    y_prev=problem.x0;
    T=4000.*ones(1,problem.nodes);
    psi=deg2rad(zeros(1,problem.nodes)); 
        for i = 1:problem.nodes
            
            epsilon_span = [(i - 1), i]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, controlRet(i), tf, problem, psi(i), T(i)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
        end
    
    t_final = t_combined*t_rif%[s]
    y_final = y_combined %[/]
    
    obj = feval(@costt, tf, y_final(end, 7));

    
solutionOut.time = problem.npts * tf; % Tempo totale
solutionOut.control = controlRet; % Controlli ottimizzati
solutionOut.state = y_final(:, 1:problem.ns); % Stati
solutionOut.objective = obj; % Obiettivo finale senza integrando
solutionOut.solTime = solTime; % Tempo di soluzione totale
end      

function obj = objectivee(input, problem)
global t_rif
    phi = input(problem.nuIndex);
    tf = input(problem.ntIndex);
    t_combined = [];
    y_combined = [];
    
    % Loop per l'integrazione su ogni intervallo tspan
    tspan=tf/problem.nodes;
    y_prev=problem.x0;
    T=4000.*ones(1,problem.nodes);
    %phi=deg2rad(phi);
    psi=deg2rad(zeros(1,problem.nodes)); 
        for i = 1:problem.nodes
            
            epsilon_span = [(i - 1), i]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, phi(i), tf, problem, psi(i), T(i)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
        end
    
    t_final = t_combined*t_rif;%[s]
    y_final = y_combined; %[/]
    
    obj = feval(@costt, tf, y_final(end, 7));
    
end

function obj = costt(tf, xf)
obj = -xf;
end


function dydt = satelliteODEE(epsilon, y, phi, tf, problem, psi, T) % satelliteODE(epsilon [/], z[/], T(i) [N], phi(i) [/], psi(i)[/], tspan[s])

    g_moon=1.623; %m/s^2
    c=290*9.80665;       %m/s
       % Adimensionalizzazione
    r_rif = 15000; %m
    v_rif = 1000;  %m/s
    m_rif = 1600;  %kg
    a_rif = v_rif^2./ r_rif; %m/s^2  =186
    g_moon_adim = g_moon / a_rif; %[/] =0.0087
    c_adim=c/v_rif;
    t_rif=(r_rif/v_rif); %s
    T= T./ (m_rif*a_rif); %N= 0.0179
  
    dydt = zeros(7, 1);
    dydt(1) = y(4);                       % Equazione per dx1/dt; Vx_adim
    dydt(2) = y(5);                       % Equazione per dx2/dt; Vy_adim
    dydt(3) = y(6);
    dydt(4) = (T/y(7))*cos(phi)*cos(psi); % a_x_adim 
    dydt(5) = (T/y(7))*cos(phi)*sin(psi); % a_y_adim
    dydt(6) = (T/y(7))*sin(phi)-g_moon_adim; % a_z_adim
    dydt(7) = -T/c_adim;               % Equazione per dx7/dt; delta_massa [kg/s]

end


function [c, ceq] = constraintss(input, problem)
    global t_rif
    phi = input(problem.nuIndex);
    tf = input(problem.ntIndex);
    t_combined = [];
    y_combined = [];
    
    % Loop per l'integrazione su ogni intervallo tspan
    tspan=tf/problem.nodes;
    y_prev=problem.x0;
    T=4000.*ones(1,problem.nodes);
    %phi=deg2rad(phi);
    psi=deg2rad(zeros(1,problem.nodes)); 
        for i = 1:problem.nodes
            
            epsilon_span = [(i - 1), i]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, phi(i), tf, problem, psi(i), T(i)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
        end
    
    t_final = t_combined*t_rif;%[s]
    y_final = y_combined; %[/]
    
    c=[];
    ceq(1) = y_final(end, 3);
    ceq(2) = y_final(end, 6);
end